#* 
#* ------------------------------------------------------------------
#* TaxForms.tcl - Tax Form code
#* Created by Robert Heller on Sat Jan 27 15:50:28 2007
#* ------------------------------------------------------------------
#* Modification History: $Log: headerfile.text,v $
#* Modification History: Revision 1.1  2002/07/28 14:03:50  heller
#* Modification History: Add it copyright notice headers
#* Modification History:
#* ------------------------------------------------------------------
#* Contents:
#* ------------------------------------------------------------------
#*  
#*     Generic Project
#*     Copyright (C) 2005  Robert Heller D/B/A Deepwoods Software
#* 			51 Locke Hill Road
#* 			Wendell, MA 01379-9728
#* 
#*     This program is free software; you can redistribute it and/or modify
#*     it under the terms of the GNU General Public License as published by
#*     the Free Software Foundation; either version 2 of the License, or
#*     (at your option) any later version.
#* 
#*     This program is distributed in the hope that it will be useful,
#*     but WITHOUT ANY WARRANTY; without even the implied warranty of
#*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*     GNU General Public License for more details.
#* 
#*     You should have received a copy of the GNU General Public License
#*     along with this program; if not, write to the Free Software
#*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#* 
#*  
#* 

package require snit
package require Tk
package require BWidget
package require Dialogs 1.0
#********************************************************************
# Tax form types, widgets, and support code
#********************************************************************

namespace eval TaxForms {
  #********************************************
  # Basic tax form.  This is just a BWidget NoteBook widget in a frame
  # Associated with this NoteBook is a vector of form elements (tax lines)
  # and some support variable and options
  #********************************************
  snit::widget TaxForm {
    # Redundent declarations -- widget class and hull type
    widgetclass TaxForm
    hulltype frame
    # Type variables:
    typevariable _AllForms {}		;# A list of all loaded forms
    typevariable _BinaryPath		;# Path to binary files
    typevariable _ExecutableFileTypes	;# Executable file types
    typevariable _ExecutableExtension	;# Executable extension
    # Options:
    option -filename -readonly yes	;# Loaded file name 
    option -solverprogram -readonly yes	;# Program to solve form
    option -notebook -readonly yes	;# Parent notebook
    option -myname -readonly yes	;# My page name in the parent notebook
    # Components
    component pages			;# My notebook of pages
    component elementvector		;# My form elements (taxlines)
    # Delegations:
    delegate option * to hull		;# Delegate all other options to the hull
    # Delegate most methods to my pages
    delegate method * to pages except {delete getframe insert}
    # Instance variables:
    variable title			;# My title.
    #*****************************
    # Type constructor -- initialize _BinaryPath and _BinaryFileTypes
    #*****************************
    typeconstructor {
      set path [info nameofexecutable]
      set _BinaryPath [file dirname $path]
      set _ExecutableExtension [file extension $path] 
      set _ExecutableFileTypes [list \
	[list {Executable Files} $_ExecutableExtension ] \
	[list {All Files} * ]]
    }
    #*****************************
    # Constructor -- create the form notebook and populate it with taxlines
    #*****************************
    constructor {args} {
      # Install our notebook
      install pages using NoteBook $win.pages -arcradius 5 -tabbevelsize 5
      pack $pages -expand yes -fill both
      $self configurelist $args		;# Configure options
      # Read form file and populate the pages.
      install elementvector using ::TaxForms::ElementVector %AUTO% \
			-notebook $win.pages -filename "$options(-filename)"
      # Add myself to the list.
      # Fetch title line and set my title.
      set title [$elementvector formtitle]
      lappend _AllForms $self
    }
    #*****************************
    # Destructor -- remove myself from the list.
    #*****************************
    destructor {
      set index [lsearch -exact $_AllForms $self]
      if {$index == 0} {
	set _AllForms [lrange $_AllForms 1 end]
      } elseif {$index > 0} {
	set _AllForms [lreplace $_AllForms $index $index]
      }
    }
    # Return my title
    method gettitle {} {return "$title"}
    # Return my title
    method settitle {newtitle} {
      set title "$newtitle"
      $elementvector setformtitle "$newtitle"
      set pagetitle "$title"
      if {[string length "$pagetitle"] > 25} {set pagetitle "[string range "$title" 0 47]..."}
      $options(-notebook) itemconfigure $options(-myname) -text "$pagetitle"
    }
    # Raise my page
    method raiseme {} {
      $options(-notebook) raise $options(-myname)
    }
    # Save tax file
    method saveme {savefile} {
      if {[$elementvector saveme "$savefile"]} {
        set options(-filename) "$savefile"
      }
    }
    # Compute taxes (run solver program)
    method computemytaxes {} {
      if {[string equal "$options(-solverprogram)" {}]} {
	set options(-solverprogram) [tk_getOpenFile \
			-defaultextension "$_ExecutableExtension" \
			-filetypes $_ExecutableFileTypes \
			-initialdir $_BinaryPath \
			-parent $win \
			-title "Select tax solver program"]
	if {[string equal "$options(-solverprogram)" {}]} {return}        
      }
      $elementvector computemytaxes "$options(-solverprogram)"
    }
    # Print tax files
    method printtaxes {} {
      $elementvector printtaxes
    }
    # Typemethod to find a form given a pagename
    typemethod findform {page} {
#      puts stderr "*** $type findform $page"
      foreach form $_AllForms {
#	puts stderr "*** $type findform: form = $form, $form cget -myname = [$form cget -myname]"
	if {[string equal "[$form cget -myname]" "$page"]} {return $form}
      }
      return  {}
    }
    # Typemethod to create a form from a file
    typemethod createformfromfile {filename notebook solverprogram} {
      # Check if file is readable
      if {![file readable "$filename"]} {
	tk_messageBox -type ok -icon error -message "File is not readable: $filename!"
	return
      }
      # Create page name
      set rootname [string tolower [file rootname [file tail "$filename"]]]
      # Create the page
      if {[catch {$notebook insert end  "$rootname" -text "$rootname"} page]} {
	tk_messageBox -type ok -icon error -message "Could not create page: $page"
	return
      }
      # Create the form
      set form [$type create "$page.taxform__%AUTO%" -filename "$filename" \
					   -solverprogram "$solverprogram" \
					   -notebook $notebook \
					   -myname "$rootname"]
      pack $form -expand yes -fill both
      # Set our title
      set title [$form gettitle]
      set pagetitle "$title"
      if {[string length "$pagetitle"] > 25} {set pagetitle "[string range "$title" 0 47]..."}
      $notebook itemconfigure $rootname -text "$pagetitle"
      # Return the form.
      return "$form"
    }
  }
  #**************************************
  # Value element -- a item associated with a tax line
  # Holds low-level form element information
  #**************************************
  snit::type value_element {
    # Options:
    option -kind -readonly yes -default {}	;# The kind of value element
    option -value -default {}			;# The value of the element
    option -column -readonly yes -default 0	;# The source file column
    option -linenum -readonly yes -default 0		;# The source file linenum
    option -parent -readonly yes -default {}	;# The tax line the value relates to
    # Instance variables:
    variable widget {}				;# The widget (if any) associated with this value
    #**************************************
    # Constructor -- just configure our options and add self to the taxline's list
    #**************************************
    constructor {args} {
      $self configurelist $args
      catch {$options(-parent) addvalue $self}
    }
    #**************************************
    # Destructor -- remove self from the taxline
    #**************************************
    destructor {
      catch {$options(-parent) removevalue $self}
    }
    # Methods to access the widget instance variables
    method setwidget {w} {set widget $w}
    method getwidget {}  {return $widget}
    # Method to incr linenum
    method incrlinenum {amount} {
      incr options(-linenum) $amount
    }
  }
  #**************************************
  # Taxline record -- one of these is created for each tax line item
  #**************************************
  snit::type taxline_record {
    # Type variables:
    typevariable _widgetIndex 0			;# Every increasing index for (unique) widget names
    # Options:
    option -linename -readonly yes -default {}	;# The line name
    option -linenum -readonly yes -default 0			;# The line number
    option -parent -readonly yes -default {}	;# The parent vector
    option -formatoffset -default 0		;# The format offset
    # Instance variables:
    variable values {}				;# The value list for this taxline
    #**************************************
    # Constructor -- configure our options and add self to our parent's list
    #**************************************
    constructor {args} {
      $self configurelist $args
      catch {$options(-parent) addtaxline $self}
    }
    #**************************************
    # Destructor -- remove self from our parent's list
    #**************************************
    destructor {
      catch {$options(-parent) removetaxline $self}
    }
    #**************************************
    # Method to add a value to our value list
    #**************************************
    method addvalue {value} {
      lappend values $value
    }
    #**************************************
    # Method to remove a value from our value list
    #**************************************
    method removevalue {value} {
      set index [lsearch $values $value]
      if {$index < 0} {
	return
      } elseif {$index == 0} {
	set values [lrange $values 1 end]
      } else {
	set values [lreplace $values $index $index]
      }
    }
    #**************************************
    # Method to check for missing entries.
    # Makes sure every value line as a form box.
    #**************************************
    method checkmissing {} {
#      puts stderr "*** $self checkmissing: (before) values = $values"
      foreach vr $values {
	set kind [$vr cget -kind]
	if {[lsearch {FLOAT TEXT INT} $kind] >= 0} {return}
      }
      if {[string length "$options(-linename)"] > 0} {
	set newitem [TaxForms::value_element %AUTO% -kind TEXT -parent $self \
						    -value {} -column 0 \
						    -linenum $options(-linenum)]
	set index [lsearch $values $newitem]
	if {$index > 0} {
	  set values [lreplace $values $index $index]
	  set values [lreplace $values 0 0 $newitem [lindex $values 0]]
	}
      }
#      puts stderr "*** $self checkmissing: (after) values = $values"
    }
    #**************************************
    # Method to insure that the line offsets are set properly for multiline
    # tax elements.
    #**************************************
    method checkoffset {linecnt offset} {
      foreach vr $values {
        set newlinecnt [expr {[$vr cget -linenum] + $offset}]
	if {$linecnt < $newlinecnt} {
	  set linecnt $newlinecnt
	}
      }
      return $linecnt
    }
    #**************************************
    # Method to make sure every line has a comment field
    #**************************************
    method checkcomments {} {
      set ncomments 0
      if {[llength $values] > 0} {
	set lineno [[lindex $values 0] cget -linenum]
      }
#      puts stderr "*** $self checkcomments (before) values = $values"
      foreach vr $values {
#	puts stderr "*** $self checkcomments: $vr cget -linenum = [$vr cget -linenum], lineno = $lineno"
	if {[$vr cget -linenum] != $lineno} {
#	  puts stderr "*** $self checkcomments: ncomments = $ncomments"
	  if {$ncomments == 0} {
	    set newv [TaxForms::value_element %AUTO% -kind COMMENT \
						     -parent $self \
						     -value {} -column 50 \
						     -linenum $lineno]
	    set nidx [lsearch $values $newv]
	    set beforeidx [lsearch $values $vr]
#	    puts stderr "*** $self checkcomments: vr = $vr, newv = $newv, nidx = $nidx, beforeidx = $beforeidx"
#	    puts stderr "*** $self checkcomments: (before update) values = $values"
	    set values [lreplace $values $nidx $nidx]
	    set values [lreplace $values $beforeidx $beforeidx $newv $vr]
#	    puts stderr "*** $self checkcomments: (after update) values = $values"
	  }
	  set ncomments 0
	}
	if {[string equal [$vr cget -kind] COMMENT]} {incr ncomments}
	set lineno [$vr cget -linenum]
      }
      if {$ncomments == 0} {
	TaxForms::value_element %AUTO% -kind COMMENT -parent $self \
                                                     -value {} -column 50 \
                                                     -linenum $lineno
      }
#      puts stderr "*** $self checkcomments (after) values = $values"
    }
    #**************************************
    # Method to display a tax line element and all of its widgets
    #**************************************
    method placeformdata {offset} {
#      puts stderr "*** $self placeformdata $offset"
#      puts stderr "*** $self placeformdata: options(-formatoffset) = '$options(-formatoffset)'"
      # Compute the row (linenum) and page for the linename.
      incr offset $options(-formatoffset)
      set linoff [expr {$options(-linenum) + $offset}]
      set linesperpage [TaxForms::ElementVector linesperpage]
      set pagenum [expr {$linoff / $linesperpage}]
      set linenum [expr {$linoff % $linesperpage}]
#      puts stderr "*** $self placeformdata: ($options(-linename)) linenum = $linenum"
      set pagept [$options(-parent) pagept $pagenum]
      # Place the line label (col 0)
      set olinenum $linenum
      # Insure unique widget names
      incr _widgetIndex
      grid [Label::create $pagept.linename$_widgetIndex -text "$options(-linename)" \
						   -anchor w] \
	-column 0 -row $linenum -sticky nws
      set lastline -1
      set button {}
      set capgtoggle no
#      puts stderr "*** $self placeformdata: values = $values"
      set ccol 1 ;# If comment only, start it in column 1 (and span to column 3)
      foreach vr $values {
	# Compute row (linenum) and page for this value item
	set linoff [expr {[$vr cget -linenum] + $offset}]
	set pagenum [expr {$linoff / $linesperpage}]
	set linenum [expr {$linoff % $linesperpage}]
#	puts stderr "*** $self placeformdata: vr = $vr ([$vr cget -kind]: [$vr cget -value]), linenum = $linenum"
	set pagept [$options(-parent) pagept $pagenum]
	# Insure unique widget names
	incr _widgetIndex
	switch [$vr cget -kind] {
	  INT -
	  FLOAT {
	    # Ints and floats are just vanila Entry widgets
	    set w [Entry::create $pagept.value$_widgetIndex \
					-text "[$vr cget -value]"]
	    grid $w -column 1 -row $linenum -sticky news
	    $vr setwidget $w
	    # Add another box button
	    set button [Button::create $pagept.addbox$_widgetIndex -text {+} \
				-padx 0 -pady 0 \
				-command [mymethod addnewbox $vr]]
	    grid $button -column 2 -row $linenum -sticky s
	    set lastline $linenum
	    set ccol 3
	  }
	  TEXT {
	    # Text entries are more complicated...
	    # Check for Capital gains or Status.  
	    # Use a LabelEntry for Capital gains and a ComboBox for Status,
	    # otherwise, just a plain Entry.
	    # Include a proper add box button. 
	    if {$lastline == $linenum} {
	      # Cap-gains.
	      if {![string equal "$button" {}]} {
		destroy $button
		set button {}
	      }
	      if {!$capgtoggle} {
		set w [LabelEntry::create $pagept.value$_widgetIndex \
					-label "Date Bought" -labelwidth 12 \
					-text "[$vr cget -value]"]
	      } else {
		set w [LabelEntry::create $pagept.value$_widgetIndex \
					-label "Date Sold" -labelwidth 12 \
					-text "[$vr cget -value]"]

		grid $w -column 1 -row $linenum -sticky news
	      }
	      set capgtoggle [expr {! $capgtoggle}]
	      set button [Button::create $pagept.addbox$_widgetIndex -text {+} \
				-padx 0 -pady 0 \
				-command [mymethod addnewgcbox $vr]]
	      grid $button -column 2 -row $linenum -sticky s
	    } elseif {[string equal -nocase "$options(-linename)" "status"]} {
	      set w [ComboBox::create $pagept.value$_widgetIndex \
				-editable no -values {Single Married/Joint 
						      Married/Sep 
						      Head_of_Household 
						      Widow(er)} \
				-text "[$vr cget -value]"]
	      grid $w -column 1 -row $linenum -sticky news
	    }  else {
	      set w [Entry::create $pagept.value$_widgetIndex \
					-text "[$vr cget -value]"]
	      grid $w -column 1 -row $linenum -sticky news
	      set button [Button::create $pagept.addbox$_widgetIndex -text {+} \
				-padx 0 -pady 0 \
				-command [mymethod addnewbox $vr]]
	      grid $button -column 2 -row $linenum -sticky s
	    }
	    $vr setwidget $w
	    set lastline $linenum
	    set ccol 3
	  }
	  COMMENT {
	    # Comments.  Use a Label and add an edit comment button
	    set cl [Label::create $pagept.comment$_widgetIndex \
					-text "[$vr cget -value]"]
	    # Check if this is a long comment or not.
	    if {$ccol == 3} {
	      grid $cl -column 3 -row $linenum -sticky w
	    } else {
	      grid $cl -column 1 -columnspan 3 -row $linenum -sticky w
	    }
	    # Edit comment button.
	    set cb [Button::create $pagept.editcomment$_widgetIndex \
				-text {*} -padx 0 -pady 0 \
				-command [mymethod editlinecomment $vr]]
	    grid $cb -column 4 -row $linenum -sticky s
	  }
	}
      }
      # Pass offset back for next line.
      return $offset
    }
    #**************************************
    # Edit a comment line
    #**************************************
    method editlinecomment {item} {
      set currentComment "[$item cget -value]"
      foreach {button result} [Dialogs::EditStringDialog draw \
					-currentstring "$currentComment" \
					-title "Edit Line Comment" \
					-parent [$options(-parent) cget \
								-notebook] \
					] {break}
      switch $button {
	ok {
	  $item configure -value "$result"
	  $options(-parent) updateboxinfo
	  $options(-parent) refresh
	}
      }
    }
    #**************************************
    # Add one new box
    #**************************************
    method addnewbox {item} {
      $self addboxes $item 1
    }
    #**************************************
    # Add two new boxes (Capital Gains)
    #**************************************
    method addnewgcbox {item} {
      $self addboxes $item 2
    }
    #**************************************
    # Callback for the "+" button on form-boxes.
    # Adds new form-box(es) to the line item.
    #**************************************
    method addboxes {item num} {
      $options(-parent) updateboxinfo
      set newitem1 [TaxForms::value_element \
			%AUTO% -kind TEXT -parent $self \
			       -value {} -column 0 \
			       -linenum [expr {[$item cget -linenum] + 1}]]
      if {$num > 1} {
	set newitem2 [TaxForms::value_element \
			%AUTO% -kind TEXT -parent $self \
			       -value {} -column 0 \
			       -linenum [expr {[$item cget -linenum] + 2}]]
      } else {
	set newitem2 $newitem1
      }
      # Skip the items on the original line
      set lineitemIndex [lsearch $values $item]
      set curline [$item cget -linenum]
      incr lineitemIndex
      set lineitemNext [lindex $values $lineitemIndex]
      while {![string equal $lineitemNext $newitem1] &&
	     [$lineitemNext cget -linenum] == $curline} {
	incr lineitemIndex
	set lineitemNext [lindex $values $lineitemIndex]
      }
      # Insert in list and remove from tail, if not at end of list.
      if {![string equal $lineitemNext $newitem1]} {
	set tail [lsearch $values $newitem1]
	incr tail -1
	set values [lrange $values 0 $tail]
	if {$num == 1} {
	  set values [lreplace $values $lineitemIndex $lineitemIndex \
					$newitem1 $lineitemNext]
	} else {
	  set values [lreplace $values $lineitemIndex $lineitemIndex \
					$newitem1 $newitem2 $lineitemNext]
	}
      }
      # Increment the effective file-line-number of all subseqent line entries.
      # First the remaining (if any) lines of this item
      set lineitemIndex [lsearch $values $newitem2]
      incr lineitemIndex
      while {$lineitemIndex < [llength $values]} {
	set lineitemNext [lindex $values $lineitemIndex]
	$lineitemNext incrlinenum $num
	incr lineitemIndex
      }
      # Next do the remaining line items
      $options(-parent) incrlinenumsafter $self $num
      $options(-parent) refresh
    }
    #**************************************
    # Incrment my linenums
    #**************************************
    method incrlinenums {amount} {
      incr options(-linenum) $amount
      foreach item $values {
	$item incrlinenum $amount
      }
    }
    #**************************************
    # Update box info
    #**************************************
    method updatemyboxes {} {
      foreach item $values {
	set w [$item getwidget]
	if {[string equal "$w" {}]} {continue}
	$item configure -value "[$w cget -text]"
      }
    }
    #**************************************
    # Write out taxline
    #**************************************
    method writetaxline {fp} {
      puts -nonewline $fp "$options(-linename)"
      set semicolon no
      set lastline $options(-linenum)
      foreach item $values {
	if {[$item cget -linenum] != $lastline} {
	  puts $fp {}
	  set lastline [$item cget -linenum]
	}
	switch [$item cget -kind] {
	  FLOAT {
	    puts -nonewline $fp "[format {	%6.2f	} [$item cget -value]]"
	  }
	  INT {
	    puts -nonewline $fp "[format {	%d	} [$item cget -value]]"
	  }
	  TEXT {
	    puts -nonewline $fp "[format {	%s	} [$item cget -value]]"
	  }
	  COMMENT {
	    set v "[$item cget -value]"
	    if {[string length "$v"] > 0} {
	      puts -nonewline $fp "	\{$v\}"
	    }
	  }
	  COLON {
	    set semicolon yes
	  }
	}
      }
      puts $fp {}
      if {$semicolon} {
	puts $fp {	;}
      }
    }
  }
  #**************************************
  # Element Vector type -- holds the vector (list) of tax lines
  #**************************************
  snit::type ElementVector {
    # Type variables:
    typevariable _LinesPerPage 10	;# *CONSTANT* Number of lines per page
    # Type method to return the number of lines per page.
    typemethod linesperpage {} {return $_LinesPerPage}
    # Options:
    option -notebook -readonly yes	;# Our notebook
    option -filename -readonly yes	;# Loaded file name
    # Instance variables
    variable linebuffer {}		;# Line buffer
    variable ots_column -1		;# Current line column index
    variable ots_linenum 0		;# Current line number index
    variable taxlines {}		;# Tax line list
    variable panels -array {}		;# Panel array
    variable title {}			;# Form title
    variable outputfile {}		;# Output file (solved taxes)
    #**************************************
    # Constructor -- read in Tax Data File and build taxline list.
    #**************************************
    constructor {args} {
      $self configurelist $args
      set state 0
      set column 0
      set linenum 0
      set linecnt 0
      set lastline 0
      set newentry no
      set txline {}
      # Open file
      set fp [open "$options(-filename)" r]
      # Fetch title line -- title must be the first line!
      set titleline [gets $fp]
      regsub {^Title:[[:space:]]*} "$titleline" {} title
      # Get first entry
      set kind [$self getNextEntry word column linenum $fp]
#      puts stderr "*** $type create $self: kind = $kind, word = $word, column = $column, linenum = $linenum"
      if {$linenum > $lastline} {
	set lastline $linenum
	if {$newentry} {
	  incr linecnt
	  set newentry no
	}
      }
      while {![eof $fp]} {
	if {$column == 0} {set state 0}
	switch -exact $kind {
	  VALUE_LABEL {
	    if {$state == 0} {
#	      puts stderr "*** $type create $self: LineLabel: $word ($linecnt)"
	      set state 1
	      set txline [TaxForms::taxline_record create %AUTO% \
				-linename "$word" \
				-linenum $linecnt \
				-parent $self]
	    } else {
#	      puts stderr "*** $type create $self: Value: $word ($linecnt)"
	      if {[string equal -nocase"[$txline cget -linename]" "status"]} {
		TaxForms::value_element %AUTO% -kind TEXT -parent $txline \
					       -value "$word" -column $column \
					       -linenum $linecnt
		set state 0
	      } else {
		TaxForms::value_element %AUTO% -kind TEXT -parent $txline \
					       -value "$word" -column $column \
					       -linenum $linecnt
	      }
	    }
	    set newentry yes
	  }
	  COMMENT {
#	    puts stderr "*** $type create $self: Comment: $word ($linecnt)"
	    if {[string equal "$txline" {}]} {
	      set txline [TaxForms::taxline_record create %AUTO% \
					-linename {} -linenum $linenum \
					-parent $self]
	    }
	    TaxForms::value_element %AUTO% -kind COMMENT -parent $txline \
						-value "$word" -column $column \
                                                -linenum $linecnt
	    set newentry yes
	  }
	  SEMICOLON {
#	    puts stderr "*** $type create $self: End: $word"
	    set state 0
	    TaxForms::value_element %AUTO% -kind COLON -parent $txline \
						-value "$word" -column $column \
						-linenum $linecnt
	  }
	}
	#incr column [string length "$word"] ;# Useless line
	set kind [$self getNextEntry word column linenum $fp]
#        puts stderr "*** $type create $self: kind = $kind, word = $word, column = $column, linenum = $linenum"
	if {$linenum > $lastline} {
	  set lastline $linenum
	  if {$newentry} {
	    incr linecnt
	    set newentry no
	  }
	}
      }
      close $fp
      # Check for missing entries.
      foreach txline $taxlines {
	$txline checkmissing
      }
      $self displaytaxinfo
    }
    #**************************************
    # Return form title.
    #**************************************
    method formtitle {} {return "$title"}
    #**************************************
    # Set form title.
    #**************************************
    method setformtitle {newtitle} {set title "$newtitle"}
    #**************************************
    # Return page frame given page index.
    #**************************************
    method pagept {page} {
      if {[catch {set panels($page)} pagept]} {
	error "Too many pages!"
      } else {
	return $pagept
      }
    }
    #**************************************
    # Method to add a taxline to the list of taxlines.
    #**************************************
    method addtaxline {taxline} {
      lappend taxlines $taxline
    }
    #**************************************
    # Method to remove a taxline from the list of taxlines.
    #**************************************
    method removetaxline {taxline} {
      set index [lsearch $taxlines $taxline]
      if {$index < 0} {
	return
      } elseif {$index == 0} {
	set taxlines [lrange $taxlines 1 end]
      } else {
	set taxlines [lreplace $taxlines $index $index]
      }
    }
    #**************************************
    # Method to read next item from the input file.
    # Returns NOTHING on EOF
    # Returns COMMENT for a comment
    # Returns VALUE_LABEL for a data item or line label
    # Returns SEMICOLON for a ';' end-of-entry character
    #
    # Passes back the string, the column, and line number into the 
    # variables named wordvar, columnvar, and linenumvar.  Fp is the
    # open channel to read from.
    #**************************************
    method getNextEntry {wordvar columnvar linenumvar fp} {
      upvar $wordvar word
      upvar $columnvar column
      upvar $linenumvar linenum

      # Skip leading whitespace.
      incr ots_column
      while {[string length [string trim [string range \
						 "$linebuffer" \
						 $ots_column end]]] == 0} {
	set linebuffer [gets $fp]
	set ots_column 0
	incr ots_linenum
	if {[eof $fp]} {break}
      }
      set ots_column [string first \
			     [string trimleft \
				     [string range "$linebuffer" \
						    $ots_column end]] \
			     "$linebuffer" $ots_column]
      # Get first non-whitespace character.
      set word [string index "$linebuffer" $ots_column]
      # Save starting column and line number
      set column $ots_column
      set linenum $ots_linenum
      # Return NOTHING on EOF
      if {[eof $fp]} {
	set word {}
	return NOTHING
      }
      if {[string equal "$word" "\{"]} {
	# Get_Comment
	set ots_column [string first "\}" "$linebuffer" $column]
#	puts stderr "*** $self getNextEntry: column = $column, ots_column = $ots_column"
	if {$ots_column > $column} {
	  set word [string range "$linebuffer" [expr {$column + 1}] \
					       [expr {$ots_column - 1}]]
#	  puts stderr "*** $self getNextEntry: word = '$word'"
	  return COMMENT
	} else {
	  set word [string range "$linebuffer" [expr {$column + 1}] end]
#	  puts stderr "*** $self getNextEntry: word = '$word'"
	  if {[eof $fp]} {return COMMENT}
	  set linebuffer "[gets $fp]"
	  set ots_column 0
	  incr ots_linenum
	  set ots_column [string first "\}" "$linebuffer"]
#	  puts stderr "*** $self getNextEntry: ots_column = $ots_column"
	  while {$ots_column < 0} {
	    append word "\n$linebuffer"
#	    puts stderr "*** $self getNextEntry: word = '$word'"
	    set ots_column 0
	    incr ots_linenum
	    if {[eof $fp]} {return COMMENT}
	    set linebuffer "[gets $fp]"
	    set ots_column [string first "\}" "$linebuffer"]
	  }
	  append word "\n[string range $linebuffer 0 [expr {$ots_column - 1}]]"
#	  puts stderr "*** $self getNextEntry: word = '$word'"
	  return COMMENT
        }
      } else {
	# Get_Value_or_Linelabel
	for {set k 1} {true} {incr k} {
	  set index [expr {$column + $k}]
	  if {$index >= [string length "$linebuffer"]} {break}
	  set ch [string index "$linebuffer" $index]
	  if {[string is space "$ch"]} {break}
	  if {[string equal "$ch" {;}]} {break}
	  append word "$ch"
	  incr ots_column
	}
	if {[string equal "$word" {;}]} {
	  return SEMICOLON
	} else {
	  return VALUE_LABEL
	}
      }
    }
    method checkcomments {} {
      foreach txline $taxlines {
	$txline checkcomments
      }
    }
#*************************************************************************
#* Display the Tax Info - This routine constructs, lays-out and poulates *
#*  the panels.  Called after initial read-in and on updates.		 *
#*************************************************************************
    method displaytaxinfo {{currentPage page1}} {
      $self checkcomments
      # First count the number of lines to display
      set linecnt 0
      set offset 0
      foreach txline $taxlines {
	if {$linecnt < [expr {[$txline cget -linenum] + $offset}]} {
	  set linecnt [expr {[$txline cget -linenum] + $offset}]
	}
        set linenum [expr {$linecnt % $_LinesPerPage}]
	set olinecnt $linecnt
	set linecnt [$txline checkoffset $linecnt $offset]
	# Don't start a new multi-line entry at the bottom of a page
	if {$linenum == [expr {$_LinesPerPage-1}] && $linecnt > $olinecnt} {
	  $txline configure -formatoffset 1
	  incr offset
	  incr linecnt
	} else {
	  $txline configure -formatoffset 0
	}
      }
      set nlines $linecnt
      # Now create enough tabbed-panels (notebook pages) to hold all of the lines.
      set npanels [expr {$nlines / $_LinesPerPage + 1}]
      for {set page 0} {$page < $npanels} {incr page} {
	set pagenum [expr {$page + 1}]
	set name page$pagenum
        set tab  "Page $pagenum"
	set panels($page) [$options(-notebook) insert end $name -text "$tab"]
      }
      set offset 0
      # Now place the form data on the pages (all work pushed into taxline object -- OO is FUN!)
      foreach txline $taxlines {
	set offset [$txline placeformdata $offset]
      }
      # Add the advance and go-back paging buttons to the bottom of the pages
      for {set page 0} {$page < $npanels} {incr page} {
	set pagept $panels($page)
	set nextpagename page[expr {$page + 2}]
	set prevpagename page$page
	if {$page > 0} {
	  set back [ArrowButton::create $pagept.back \
		-dir left -type button \
		-command "$options(-notebook) raise $prevpagename"]
	  grid $back -column 0 -row $_LinesPerPage -sticky sw
	}
	if {$page < [expr {$npanels -1}]} {
	  set back [ArrowButton::create $pagept.next \
		-dir right -type button \
		-command "$options(-notebook) raise $nextpagename"]
	  grid $back -column 4 -row $_LinesPerPage -sticky se
	}	
      }
      $options(-notebook) raise $currentPage
    }
    #**************************************
    # Method to update line numbers in taxlines
    #**************************************
    method incrlinenumsafter {afterme amount} {
      set found no
      foreach taxline $taxlines {
	if {!$found && ![string equal $taxline $afterme]} {continue}
	if {!$found} {
	  set found yes
	  continue
	} else {
	  $taxline incrlinenums $amount
	}
      }
    }
    #**************************************
    # Refresh display
    #**************************************
    method refresh {} {
      set pagenum [$options(-notebook) raise]
      foreach p [array names panels] {
	$options(-notebook) delete page[expr {$p + 1}] 1
      }
      array unset panels
      $self displaytaxinfo $pagenum
    }
    #**************************************
    # Update box info
    #**************************************
    method updateboxinfo {} {
      foreach taxline $taxlines {
	$taxline updatemyboxes
      }
    }
    #**************************************
    # Save tax file
    #**************************************
    method saveme {savefile} {
      $self updateboxinfo
      if {[catch {open "$savefile" w} fp]} {
	tk_messageBox -icon error -type ok -message "Could not open $savefile: $fp"
	return false
      }
      puts $fp "Title: $title"
      foreach taxline $taxlines {
	$taxline writetaxline $fp
      }
      close $fp
      set options(-filename) "$savefile"
      tk_messageBox -icon info -type ok -message "Write form-data to file $savefile"
      return true
    }
    #**************************************
    # Compute taxes (run solver program)
    #**************************************
    method computemytaxes {solverprogram} {
      Dialogs::SubprocessDialog draw \
			-command "$solverprogram $options(-filename)" \
			-parent $options(-notebook)
      set outputfile [file rootname "$options(-filename)"].out
      tk_messageBox -type ok -icon info -message "Results saved in $outputfile"
    }
    #**************************************
    # Print tax file
    #**************************************
    method printtaxes {} {
      set whichfile [Dialogs::SelectFileToPrintDialog draw \
			-source "$options(-filename)" \
			-output "$outputfile" \
			-parent $options(-notebook)]
      if {[string equal "$whichfile" {}]} {return}
      global tcl_platform
      switch $tcl_platform(platform) {
	macintosh -
	unix {
	  set lp [auto_execok lp]
	  set lpr [auto_execok lpr]
	  if {![string equal "$lp" {}]} {
	    Dialogs::SubprocessDialog draw -command "$lp $whichfile" \
			-parent $options(-notebook)
	  } elseif {![string equal "$lpr" {}]} {
	    Dialogs::SubprocessDialog draw -command "$lpr $whichfile" \
			-parent $options(-notebook)
	  } else {
	    tk_messageBox -type ok -icon error -message "Cannot find lp or lpr!"
	  }
	}
	windows {
	  set print [auto_execok print]
	  if {![string equal "$print" {}]} {
	    Dialogs::SubprocessDialog draw -command "$print $whichfile" \
			-parent $options(-notebook)
	  } else {
	    tk_messageBox -type ok -icon error -message "Cannot find print!"
	  }
	}
      }
    }
  }
}





package provide TaxForms 1.0
